@@ -283,10 +283,10 @@ static bool check_all_colors_known(int num_spot,
 
 /* If possible, update the equivalent CMYK color for a spot color */
 int
-update_spot_equivalent_cmyk_colors(gx_device * pdev, const gs_gstate * pgs,
+update_spot_equivalent_cmyk_colors(gx_device * pdev, const gs_gstate * pgs, const gs_color_space * pcs_in,
     gs_devn_params * pdevn_params, equivalent_cmyk_color_params * pparams)
 {
-    const gs_color_space * pcs;
+    const gs_color_space * pcs = pcs_in;
     cmm_dev_profile_t *dev_profile;
     int code;
 
@@ -303,12 +303,16 @@ update_spot_equivalent_cmyk_colors(gx_device * pdev, const gs_gstate * pgs,
         pparams->all_color_info_valid = true;
         return 0;
     }
+
+    /* If the caller apssed in NULL for the colour space, use the current colour space */
+    if (pcs_in == NULL)
+        pcs = gs_currentcolorspace_inline(pgs);
+
     /*
      * Verify that the given color space is a Separation or a DeviceN color
      * space.  If so then when check if the color space contains a separation
      * color for which we need a CMYK equivalent.
      */
-    pcs = gs_currentcolorspace_inline(pgs);
     if (pcs != NULL) {
         if (pcs->type->index == gs_color_space_index_Separation) {
             update_Separation_spot_equivalent_cmyk_colors(pdev, pgs, pcs,
@@ -368,7 +372,6 @@ typedef struct color_capture_device_s {
 static cmap_proc_gray(cmap_gray_capture_cmyk_color);
 static cmap_proc_rgb(cmap_rgb_capture_cmyk_color);
 static cmap_proc_cmyk(cmap_cmyk_capture_cmyk_color);
-static cmap_proc_rgb_alpha(cmap_rgb_alpha_capture_cmyk_color);
 static cmap_proc_separation(cmap_separation_capture_cmyk_color);
 static cmap_proc_devicen(cmap_devicen_capture_cmyk_color);
 
@@ -376,7 +379,6 @@ static const gx_color_map_procs cmap_capture_cmyk_color = {
     cmap_gray_capture_cmyk_color,
     cmap_rgb_capture_cmyk_color,
     cmap_cmyk_capture_cmyk_color,
-    cmap_rgb_alpha_capture_cmyk_color,
     cmap_separation_capture_cmyk_color,
     cmap_devicen_capture_cmyk_color
 };
@@ -425,14 +427,6 @@ cmap_cmyk_capture_cmyk_color(frac c, frac m, frac y, frac k, gx_device_color * p
     save_spot_equivalent_cmyk_color(sep_num, pparams, cmyk);
 }
 
-static void
-cmap_rgb_alpha_capture_cmyk_color(frac r, frac g, frac b, frac alpha,
-        gx_device_color * pdc, const gs_gstate * pgs, gx_device * dev,
-                         gs_color_select_t select)
-{
-    cmap_rgb_capture_cmyk_color(r, g, b, pdc, pgs, dev, select);
-}
-
 static void
 cmap_separation_capture_cmyk_color(frac all, gx_device_color * pdc,
      const gs_gstate * pgs, gx_device * dev, gs_color_select_t select,
